import datetime
import locale
from .util import Util

try:
    import openpyxl
except ImportError:
    openpyxl = None

try:
    import numpy
except ImportError:
    numpy = None


class Variable:
    """ 变量类 """

    OBJECT_PREFIX = '_@object_'
    REPEAT_PREFIX = '@repeat'

    def __init__(self, expr: str):
        self.sheet_name = None
        self.address = None
        self.address2 = None
        self.is_repeat = False
        self.has_at = False
        self.index = None
        self.expr = None
        if expr:
            self.load(expr)

    def clear(self):
        """ 清空 """
        self.sheet_name = None
        self.address = None
        self.address2 = None
        self.is_repeat = False
        self.has_at = False
        self.index = None
        self.expr = None

    def load(self, expr: str) -> 'Variable':
        """
        解析 expr 地址字符串, 返回 Variable 对象

        :param expr:  表达式, 可以是Excel地址，如： '{Sheet1!A1:B6｝', 也可以是一个Key名, 如：｛KeyName}
        :return:  如果是Excel地址，例如： address='{Sheet1!A1:B6｝', 返回tuple ('Sheet1', 'A1', 'B6').
                   如果不是Excel地址，例如： address='{KeyName}, 返回tuple (None, 'KeyName', None)
        """
        self.clear()

        if isinstance(expr, str):
            self.expr = expr
            expr = expr.strip()

            # 判断是否有大括号
            if Variable.is_var(expr):
                expr = expr[1:len(expr) - 1].strip()
                has_embrace = True
            else:
                raise ValueError('%s is not a variable, variable must start with "{" and end with "{"' % repr(expr))

            # 判断是否有 REPEAT_PREFIX
            if expr.startswith(Variable.REPEAT_PREFIX + ' '):
                self.is_repeat = True
                expr = expr[len(Variable.REPEAT_PREFIX):].strip()

            # 判断是否有 '@'
            if expr.startswith('@'):
                self.has_at = True
                expr = expr[1:].strip()

            # 判断是否有 '!'
            offset = expr.find('!')
            if offset > 0:
                self.sheet_name = expr[:offset].strip()
                expr = expr[offset + 1:].strip()
                # 判断是否有 ':'
                offset = expr.find(':')
                if offset > 0:
                    self.address2 = expr[offset + 1:]
                    self.address = expr[:offset]
                else:
                    self.address = expr
            elif has_embrace:
                self.address = expr

            # 取出 index
            if self.address.endswith(']') and self.address.find('[') > 0:
                index = self.address[self.address.find('[') + 1:]
                if index.endswith(']'):
                    index = index[:len(index)-1]
                self.index = index.strip()
                self.address = self.address[:self.address.find('[')].strip()

        return self

    @property
    def name(self):
        return self.address

    @staticmethod
    def is_var(word: str):
        """ 判断 word 是否是一个变量"""
        if isinstance(word, str) and word.startswith('{') and word.endswith('}'):
            return True
        return False


class VarData:
    SystemVars = ['index', 'indexA']
    """ 存有变量值的数据 """

    def __init__(self):
        self.path = ''

    @staticmethod
    def open(data):
        """ 打开数据 """
        if isinstance(data, dict):
            return DictVarData(data)
        elif Util.is_workbook(data):
            return ExcelVarData(data)
        elif type(data).__name__ == 'DataFrame' and hasattr(data, 'columns'):
            return DataFrameVarData(data)
        elif isinstance(data, str):
            Util.check_module(openpyxl, 'openpyxl')
            if Util.file_ext(data) == '.xlsx':
                obj = ExcelVarData(openpyxl.load_workbook(data, data_only=True))
                obj.path = Util.file_path(data)
                return obj

        raise ValueError("unknown variable data type %s" % repr(type(data)))

    # noinspection PyMethodMayBeStatic
    def _to_var(self, var):
        """ convert var to Variable object """
        if isinstance(var, Variable):
            return var
        elif isinstance(var, str):
            return Variable(var)
        raise ValueError("expect Variable but %s is found" % repr(type(var)))

    def has_var(self, var):
        """ 判断数据中是否有指定变量var的值 """
        var = self._to_var(var)
        if var.has_at:
            if var.address in VarData.SystemVars:
                return True

        raise NotImplementedError

    def var_length(self, var):
        """
        如果变量值是列表, 返回变量值列表长度.
        如果变量值不是列表，返回0
        """
        raise NotImplementedError

    def get_var(self, var, row=0):
        """
        取得有指定变量var的值

        :param var:  变量
        :param row:  (可选)当row=0, 返回第0行的值。当row>0, 返回第row行的值。
        :return: 返回对应行的变量值
        """
        var = self._to_var(var)

        # 处理特殊变量
        if var.has_at:
            if var.address == 'index':
                return str(row)
            elif var.address == 'indexA':
                return chr(ord('A') + row - 1)

        return self._get_var(var, row)

    def _get_var(self, var, row=0):
        raise NotImplementedError

    def close(self):
        pass


class ExcelVarData(VarData):
    """ 存有变量值的Excel数据 """

    def __init__(self, xlsx_obj):
        super().__init__()
        self.workbook = None
        if Util.is_workbook(xlsx_obj):
            self.workbook = xlsx_obj
            self.path = ''
        elif isinstance(xlsx_obj, str) and Util.file_ext(xlsx_obj) == '.xlsx':
            Util.check_module(openpyxl, 'openpyxl')
            self.workbook = openpyxl.load_workbook(xlsx_obj)
            self.path = Util.file_path(xlsx_obj)
        else:
            raise ValueError("expect Excel object but %s is found" % repr(type(xlsx_obj)))

    def has_sheet(self, sheet_name):
        """ 是否存在名为 sheet_name 的工作表 """
        if sheet_name:
            return sheet_name in self.workbook
        return False

    # noinspection PyMethodMayBeStatic
    def _convert_xls_date_format(self, fmt):
        """ 将 excel datetime 格式转换为 python 格式"""
        # TODO
        while fmt.find(']') > 0:
            fmt = fmt[fmt.find(']') + 1:]
        if fmt.endswith('@'):
            fmt = fmt[:len(fmt) - 1]
        if fmt.endswith(';'):
            fmt = fmt[:len(fmt) - 1]
        fmt = fmt.replace('yyyy', '%Y')
        fmt = fmt.replace('yy', '%y')
        fmt = fmt.replace('mm', '%m')
        fmt = fmt.replace('m', '%m')
        fmt = fmt.replace('dd', '%d')
        fmt = fmt.replace('d', '%d')
        fmt = fmt.replace('"', '')
        return fmt

    def has_var(self, var):
        """ 判断数据中是否有指定变量var的值 """
        var = self._to_var(var)
        if var.has_at:
            if var.address in VarData.SystemVars:
                return True

        if self.has_sheet(var.sheet_name):
            # noinspection PyBroadException
            try:
                sheet = self.workbook[var.sheet_name]
                cell = sheet[var.address]
                return cell.value is not None
            except Exception:
                return False

    def var_length(self, var):
        """
        如果变量值是列表, 返回变量值列表长度.
        如果变量值不是列表，返回0

        :param var: 变量
        :return: int
        """

        var = self._to_var(var)
        if self.has_sheet(var.sheet_name):
            # noinspection PyBroadException
            try:
                sheet = self.workbook[var.sheet_name]
                cell = sheet[var.address]
                # 统计单元格下方有数据的行数
                count = 0
                val = sheet.cell(cell.row + count + 1, cell.column).value
                while val is not None and val != '':
                    count += 1
                    val = sheet.cell(cell.row + count + 1, cell.column).value
                return count
            except Exception:
                pass
        return 0

    def _get_var(self, var, row=0):
        """
        取得有指定变量var的值

        :param var:  变量
        :param row:  (可选)当row=0, 返回第0行的值。当row>0, 返回第row行的值。
        :return: 返回对应行的变量值
        """
        var = self._to_var(var)
        if self.has_sheet(var.sheet_name):
            # noinspection PyBroadException
            try:
                sheet = self.workbook[var.sheet_name]
                cell = sheet[var.address]
                cell = sheet.cell(cell.row + row, cell.column)
                val = cell.value
                # process value
                if cell.is_date:
                    fmt = self._convert_xls_date_format(cell.number_format)
                    if isinstance(val, datetime.datetime):
                        if Util.has_chinese(fmt):
                            locale.setlocale(locale.LC_CTYPE, 'chinese')
                        val = val.strftime(fmt)
                elif isinstance(val, float):
                    val = round(val * 100) / 100  # TODO

                if var.has_at and isinstance(val, str):
                    val = Variable.OBJECT_PREFIX + val

                return val
            except Exception:
                pass

    def close(self):
        if self.path:
            self.workbook.close()
        self.path = ''


class DictVarData(VarData):
    """
    存有变量值的 dictionary 数据
    """
    def __init__(self, data):
        super().__init__()
        self.data = None
        if isinstance(data, dict):
            self.data = data
        else:
            raise ValueError("expect dict but %s is found" % repr(type(data)))

    def _get_value(self, key, default=None):
        """ get value from dictionary by key"""
        keys = key.split('.')
        d = self.data
        last_key = ''
        for k in keys:
            last_key = k
            if isinstance(d, dict) and k in d:
                d = d[k]
            elif isinstance(d, list) and Util.is_int(k) and len(d) > int(k):
                d = d[int(k)]
            else:
                return default, last_key
        return d, last_key

    def has_var(self, var):
        """
        判断数据中是否有指定变量var的值
        :param var:
        :return:
        """
        var = self._to_var(var)
        if var.has_at:
            if var.address in VarData.SystemVars:
                return True

        val, key = self._get_value(var.address)
        if val is None:
            return False

        if var.index is None:
            return val is not None
        else:
            # noinspection PyBroadException
            try:
                index = var.index
                if isinstance(val, list) or isinstance(val, tuple) or isinstance(val, str):
                    index = int(index)
                _ = val[index]
                return True
            except Exception:
                return False

    def var_length(self, var):
        """
        如果变量值是列表, 返回变量值列表长度.
        如果变量值不是列表，返回0

        :param var: 变量
        :return: int
        """
        var = self._to_var(var)
        val, key = self._get_value(var.address)
        return len(val) if isinstance(val, list) or isinstance(val, tuple) else 0

    def _get_var(self, var, row=0):
        """
        取得有指定变量var的值

        :param var:  变量
        :param row:  (可选)当row=0, 返回第0行的值。当row>0, 返回第row行的值。
        :return: 返回对应行的变量值
        """
        var = self._to_var(var)
        # 取得值和键名
        val, key = self._get_value(var.address)

        if val is None:
            return 0 if row < 0 else var.expr

        if row == 0:
            if isinstance(val, list):
                return key  # 当row=0， 返回键名
            else:
                if var.has_at and isinstance(val, str):
                    val = Variable.OBJECT_PREFIX + val
                return val
        else:
            if isinstance(val, list):
                # 当 row > 1, 返回列表项目
                val = val[row - 1] if len(val) >= row else None
                if var.has_at and isinstance(val, str):
                    val = Variable.OBJECT_PREFIX + val
                return val
            else:
                return None


class DataFrameVarData(VarData):
    """ 存有变量值的 DataFrame 数据 """
    def __init__(self, df):
        super().__init__()
        self.df = None
        if type(df).__name__ == 'DataFrame' and hasattr(df, 'columns'):
            self.df = df
        else:
            raise ValueError("expect DataFrame but %s is found" % repr(type(df)))

    # noinspection PyMethodMayBeStatic
    def _trim_quote(self, s):
        """ trim the leading and ending quote chars of the string """
        if isinstance(s, str):
            if s.startswith('"') and s.endswith('"'):
                return s[1: len(s)-1]
            elif s.startswith("'") and s.endswith("'"):
                return s[1: len(s) - 1]
        return s

    def has_var(self, var):
        """ 判断数据中是否有指定变量var的值 """
        var = self._to_var(var)
        if var.has_at:
            if var.address in VarData.SystemVars:
                return True

        # noinspection PyBroadException
        try:
            col_name = var.address
            if col_name == self.df.index.name:
                if var.index is None:
                    return True
                else:
                    index = int(var.index)
                    if index < len(self.df):
                        return True
            elif col_name in self.df.columns:
                if var.index is None:
                    return True
                else:
                    index = var.index
                    if self.df.index.dtype == numpy.int64 or self.df.index.dtype == numpy.int32:
                        index = int(index)
                    elif self.df.index.dtype == numpy.float32 or self.df.index.dtype == numpy.float64:
                        index = float(index)
                    else:
                        index = self._trim_quote(index)
                    _ = self.df.loc[index, col_name]
                    return True
        except Exception:
            pass

        return False

    def _get_var(self, var, row=0):
        """ 取得变量的值 """
        # noinspection PyBroadException
        try:
            var = self._to_var(var)
            col_name = var.address
            if col_name == self.df.index.name:
                # 如果 列名 是索引列
                if row == 0:
                    if var.index is None:
                        return col_name
                    else:
                        index = int(var.index)
                        return self.df.index[index]
                else:
                    return self.df.index[row - 1]
            elif col_name in self.df.columns:  # 如果 df 有这个列
                if row == 0:
                    if var.index is None:
                        return col_name  # 当row=0， 返回列名
                    else:
                        index = var.index
                        if self.df.index.dtype == numpy.int64 or self.df.index.dtype == numpy.int32:
                            index = int(index)
                        elif self.df.index.dtype == numpy.float32 or self.df.index.dtype == numpy.float64:
                            index = float(index)
                        else:
                            index = self._trim_quote(index)
                        return self.df.loc[index, col_name]
                else:
                    # 取得列的序号
                    col = self.df.columns.to_list().index(col_name)
                    if row - 1 < len(self.df):
                        return self.df.iloc[row - 1, col]
        except Exception:
            pass

    def var_length(self, var):
        """
        如果变量值是列表, 返回变量值列表长度.
        如果变量值不是列表，返回0
        """
        var = self._to_var(var)
        # 当 data 是 DataFrame
        col_name = var.address
        if col_name == self.df.index.name:
            return len(self.df)
        elif col_name in self.df.columns:
            return len(self.df)
        return 0



# this file datafile.py has the same filename to datafile.xlsx (but different extension)
# if xlwings addin is installed, press button "run main" in the Microsoft Excel tab 'xlwings'
# this file will be load, and the function main() will be called by xlwings.
#
import office
import pandas


def main():
    book = office.Excel()  # if xlwings addin called, get the active workbook
    sheet = book.active
    # print(sheet.name)
    # sheet['A12'].value = book.filename
    # sheet['A11'].value = sheet.name
    #
    df = sheet['A1:D6'].to_dataframe()
    sheet['A18'].value = df

    book.save()
    book.close()


if __name__ == "__main__":
    # if run in command line, call main()
    main()



# Process .docx, .xlsx, .pptx, .pdf files for office automation

from .util import Util
from .word import Word
from .excel import Excel
from .ppt import PPT
from .pdf import PDF
from .video import Video
import sys

__version__ = '0.2.5'
__author__ = "JoStudio"
__date__ = "2022/9/15"


# 打开office文件 （ .docx 或 .xlsx 或 .pptx 或 .pdf )
def open_file(filename, template=None, **kwargs):
    """
    打开office文件 （ .docx 或 .xlsx 或 .pptx 或 .pdf )

    :param filename:  文件名,
    :param template:
    :return:
    """
    ext = Util.file_ext(filename)
    if ext in ['.docx']:
        return Word(filename, template, **kwargs)
    elif ext in ['.xlsx']:
        return Excel(filename, template, **kwargs)
    elif ext in ['.pptx']:
        return PPT(filename, template, **kwargs)
    elif ext in ['.pdf']:
        return PDF(filename, template, **kwargs)
    else:
        raise ValueError('file extension % is not supported' % repr(ext))


# 逐个打印文件
def print_files(filename_list, **kwargs):
    """
    使用打印机逐个打印文件.

    :param filename_list: office文件名列表
    :param kwargs:
    :return: None
    """
    for filename in filename_list:
        try:
            w = open_file(filename)
            w.print(**kwargs)
        except Exception as e:
            print("error print file %s, %s" % (filename, str(e)), file=sys.stderr)



import office

try:
    import pandas as pd
except ImportError:
    pd = None


def example_fill_excel_data():
    # create file output.pptx using template1.pptx
    ppt = office.open_file("output.pptx", template="template1.pptx")

    # fill into the ppt with the data from Excel file datafile.xlsx, and save.
    ppt.fill('datafile.xlsx').save()


def example_fill_table():
    # create file output.pptx using template2.pptx
    # fill into the ppt with the data from Excel file datafile.xlsx, and save.
    office.open_file("output.pptx", "template2.pptx").fill('datafile.xlsx').save()


def example_fill_chart():
    # create file output.pptx using template3.pptx
    # fill into the ppt with the data from Excel file datafile.xlsx, and save.
    office.open_file("output.pptx", "template3.pptx").fill('datafile.xlsx').save()


def example_insert_photo():
    # create file output.pptx using template4.pptx
    # fill into the ppt with the data from Excel file datafile.xlsx, and save.
    office.open_file("output.pptx", "template4.pptx").fill('datafile.xlsx').save()


def example_report():
    # create file output.pptx using template5.pptx
    # fill into the ppt with the data from Excel file datafile.xlsx, and save.
    office.open_file("report.pptx", "template5.pptx").fill('datafile.xlsx').save()


def example_long_image():
    # open report.pptx, save it as a long image(save file is a .jpg file means save long imag)
    # add watermark when save
    office.open_file("report.pptx").save("long.jpg", watermark="CONFIDENTIAL")


def example_save_pdf():
    # save as .pdf file, add watermark
    office.open_file("report.pptx").save('report.pdf', watermark="商业秘密，注意保管")


def example_play_ppt():
    # play powerpoint file with timing parameters
    office.open_file("report.pptx").play([1, 3, 3, 1])


def example_play_ppt_with_voice():
    # play powerpoint file with timing parameters, play voice files
    office.open_file("report.pptx").play([
        1,
        [3, "1.wav"],
        3,
        [1, "2.wav"]
    ])


def example_play_ppt_with_file():
    # play powerpoint file with definitions in file  play.txt
    office.open_file("report.pptx").play("play.txt")


def example_fill_dictionay():
    data = {
        "Name": "Peter",
        "Age": 18,
        "Photo": "peter.jpg",
        'Study': {
            'Lesson': ['Literature', 'Math', 'English'],
            'Score': [95.3, 68, 75],
            'Evaluate': ['Good', 'Bad', 'Medium']
        }
    }

    # create file output.pptx using template5.pptx
    # fill into the ppt with the data from dictionary, and save.
    office.open_file("report.pptx", "template6.pptx").fill(data).save()


def example_fill_dataframe():
    data = {
            'Lesson': ['Literature', 'Math', 'English'],
            'Score': [95.3, 68, 75],
            'Evaluate': ['Good', 'Bad', 'Medium']
    }

    df = pd.DataFrame(data)

    # create file output.pptx using template8.pptx
    # fill into the ppt with the data from dictionary, and save.
    office.open_file("report.pptx", "template7.pptx").fill(df).save()


def example_multi_fill():
    person = {"Name": "Peter", "Age": 18, "Photo": "peter.jpg"}

    # create file output.pptx using template8.pptx
    # fill data from file datafile.xlsx, then fill data from person dict
    office.open_file("report.pptx", "template8.pptx").fill('datafile.xlsx').fill(person).save()


def example_duplicate_slide():
    # create file output.pptx using template9.pptx,  fill data from file datafile.xlsx
    office.open_file("report.pptx", "template9.pptx").fill("datafile.xlsx").save()


if __name__ == "__main__":
    example_fill_excel_data()
    # example_fill_table()
    # example_fill_chart()
    # example_insert_photo()
    # example_report()
    # example_long_image()
    # example_save_pdf()
    # example_play_ppt()
    # example_play_ppt_with_voice()
    # example_play_ppt_with_file()
    # example_fill_dictionay()
    # example_fill_dataframe()
    # example_multi_fill()
    # example_duplicate_slide()






import office


def example_basic():
    # open file datafile.xlsx, return workbook object
    workbook = office.open_file("datafile.xlsx")

    # list worksheets
    sheet_names = workbook.sheetnames
    print(sheet_names)

    # get a sheet
    sheet = workbook['Sales']
    print(sheet.name)

    # get a cell
    cell = sheet['B2']
    print(cell.value)

    # write a cell
    cell.value = 'New City'

    # save
    workbook.save("newfile.xlsx")

    # close
    workbook.close()


def example_workbook_worksheet():
    # create a newfile.xlsx, copied from the template file
    # if the newfile.xlsx not exists, create it.
    # if the newfile.xlsx already exists, overwrite it.
    workbook = office.open_file("newfile.xlsx", template="datafile.xlsx")

    # create a new sheet
    workbook.sheets.add("NewSheet")

    # if there is a sheet named "Sales"
    if 'NewSheet' in workbook:
        sheet = workbook['NewSheet']  # get the sheet
        sheet.name = 'NewName'  # change sheet name
        workbook.active = sheet  # set active sheet

    workbook.sheets.add("NewSheet2")  # create a sheet
    workbook.sheets.delete('NewSheet2')  # delete sheet

    workbook.save()
    workbook.close()


def example_read_write_cell():
    # open file datafile.xlsx, return workbook object
    workbook = office.open_file("datafile.xlsx")

    # if there is a worksheet named 'Sales'
    if 'Sales' in workbook:
        sheet = workbook['Sales']  # get the worksheet object

        print('max_row', sheet.max_row)
        print('max_column', sheet.max_column)

        cell = sheet['B3']  # get the 'B2' cell object
        print(cell.value)  # print the value of the cell

        cell = sheet.cell(3, 2)  # get cell at row 2, column 3 (B3)
        print(cell.value)  # print the value of the cell

        cell.value = 'New City'  # Change the value of the cell

    workbook.save()
    workbook.close()


def example_read_write_range():
    # open file datafile.xlsx, return workbook object
    workbook = office.open_file("datafile.xlsx")

    # if there is a worksheet named 'Sales'
    if 'Sales' in workbook:
        sheet = workbook['Sales']  # get the worksheet object
        # read a range value
        range1 = sheet['B2:C4']
        # value of the range is a list, include values of the cells
        print(range1.value)

        # change the values in the range
        sheet['B11:C12'].value = [['Chicago', 18], ['Shanghai', 20]]


def example_read_write_via_workbook():
    # open file datafile.xlsx, return workbook object
    wb = office.open_file("datafile.xlsx")

    # get a sheet by name
    sheet = wb['Sales']  # if sheet not exists, raise exception

    cell = sheet['B3']  # get a cell object
    print(cell.value)  # print the cell value
    print(sheet.cell(3, 2).value)  # get a cell by (row_int, column_int)

    # write the value of the cell
    cell.value = 'Paris'

    # get a cell through workbook, if worksheet 'Sales' not exists, create it.
    print(wb['Sales!B3'].value)
    print(wb['Sales', 'B3'].value)  # by sheet_name, cell address
    print(wb['Sales', 3, 2].value)  # by sheet_name, row_int, column_int

    # write a cell value, if worksheet 'Sales' not exists, create it.
    wb['Sales!B3'].value = "Shanghai"

    print(wb['Sales!B3'].value)

    # read a range value
    range1 = wb['Sales!B2:C4']
    # value of the range is a list, include values of the cells
    print(range1.value)

    # change the values in the range
    wb['Sales!B11:C12'].value = [['Chicago', 18], ['Shanghai', 20]]

    print(wb['Sales!B11:C12'].value)

    wb.save()
    wb.close()


def example_find():
    # open file, get a workbook object
    wb = office.open_file("datafile.xlsx")

    # get a sheet
    sheet = wb["Sales"]

    # find cell value "Sydney", return cell object. return None if not found.
    cell = sheet.find("Sydney")
    if cell is not None:  # if found
        # print info of cell
        print(cell.value, cell.address, cell.row, cell.column)

    # find two value: "Sydney", "Sales Amount"
    # return the cross cell.   return None if not found
    cell = sheet.find("Sydney", "Sales Amount")
    if cell is not None:  # if found
        # print info of cell
        print(cell.value, cell.address, cell.row, cell.column)

    # using workbook object to find
    cell = wb.find("Sales", "Sydney", "Sales Amount")
    if cell is not None:  # if found
        print("Sydney", "Sales Amount", cell.value)


def example_expand():
    # open file, get a workbook object
    wb = office.open_file("datafile.xlsx")

    # get a cell : Sales!A1
    cell = wb["Sales!A1"]

    # expand the cell to a range, result range will be A1:D6
    range1 = cell.expand()
    print(range1)

    # expand the cell to a range, do not expand to the right
    print(cell.expand(right=None))  # A1:A6

    # expand the cell to a range, do not expand downward
    print(cell.expand(down=None))  # A1:D1

    # expand rightward with specified columns count, downward with specified rows count
    print(cell.expand(right=5, down=4))  # A1:F5


def example_range_to_dataframe():
    # open file, get a workbook object
    wb = office.open_file("datafile.xlsx")

    # get a range  Sales!A1:D6
    range1 = wb["Sales", 'A1:D6']
    print(range1)

    # get a list data from the range
    list1 = range1.to_list()
    print(list1)

    # get a DataFrame object from the range
    df = range1.to_dataframe()
    print(df)

    # Write DataFrame to the range
    range1.values = df

    # Write list data to the range
    range1.values = list1

    # wb.save()
    # wb.close()


if __name__ == "__main__":
    # example_basic()
    # example_workbook_worksheet()
    # example_read_write_cell()
    # example_read_write_range()
    # example_read_write_via_workbook()
    # example_expand()
    # example_find()
    # example_expand()
    # example_range_to_dataframe()
    pass




import os

from .util import Util

try:
    from win32com.client import Dispatch, DispatchEx, CDispatch
except ImportError:
    Dispatch = None
    DispatchEx = None
    CDispatch = None

try:
    import win32gui
    import win32api
    import winsound
except ImportError:
    win32gui = None
    win32api = None
    winsound: None = None


class Win32:

    @staticmethod
    def check():
        if not DispatchEx:
            raise ImportError('win32com module is required')

    @staticmethod
    def create_com_object(clsid, exception=True):
        """ 创建 COM 对象  """
        try:
            app = DispatchEx(clsid)
            return app
        except Exception as e:
            if exception:
                raise ValueError('Cannot create win32 COM object of %s , %s' % (repr(clsid), str(e)))
        return None

    @staticmethod
    def powerpoint_open(filename):
        """
        启动 PowerPoint应用， 并打开文件

        :param filename:  文件名
        :return:  返回(app, presentation)
        """
        Win32.check()

        # https://cloud.tencent.com/developer/article/1907286
        app = Win32.create_com_object('Kwpp.Application', False)  # WPS
        if app is None:
            app = Win32.create_com_object('PowerPoint.Application')

        app.Visible = 1  # 前台运行
        app.DisplayAlerts = 0  # 不显示，不警告
        presentation = None

        # open file
        if filename:
            if filename.find(':') <= 0:
                filename = os.path.join(os.getcwd(), filename)
            presentation = app.Presentations.Open(filename)

        return app, presentation

    @staticmethod
    def powerpoint_close(app, presentation=None):
        """ 关闭 PowerPoint应用 """
        # noinspection PyBroadException
        try:
            if presentation:
                presentation.Saved = True
                presentation.Close()  # 关闭 PowerPoint 文档
        except Exception:
            pass

        # noinspection PyBroadException
        try:
            app.Quit()
        except Exception:
            pass

    @staticmethod
    def _open(clsid, command=None, filename=None, visible=True, current_dir=True):
        """
        打开应用， 打开文件
        :param filename: 文件名
        :return:  返回 tuple (app, file_obj)
        """
        Win32.check()

        clsids = clsid.split('|')
        # 打开应用
        app = Win32.create_com_object(clsids[0], False)
        if app is None and len(clsids) > 1:
            app = Win32.create_com_object(clsids[1])

        if hasattr(app, 'Visible'):
            # noinspection PyBroadException
            try:
                app.Visible = visible
            except Exception:
                pass

        if hasattr(app, 'DisplayAlerts'):
            # noinspection PyBroadException
            try:
                app.DisplayAlerts = 0  # Do not show alert (不警告)
            except Exception:
                pass

        # 打开文件
        file_obj = None
        if filename:
            if current_dir and not Util.is_absolute_path(filename):
                filename = os.path.join(os.getcwd(), filename)
            if hasattr(app, command):
                files = getattr(app, command)
                file_obj = files.Open(filename)
            else:
                raise ValueError('cannot find command %s' % repr(command))

        return app, file_obj

    @staticmethod
    def open_word(filename, visible=True):
        """
        启动 Word应用， 并打开文件

        :param filename:  文件名
        :param visible:  是否显示应用界面
        :return:  返回 (app, document)
        """
        # https://cloud.tencent.com/developer/article/1907286
        return Win32._open('Word.Application|KWPS.Application', 'Documents', filename, visible)

    @staticmethod
    def open_excel(filename, visible=True):
        """
        启动 Excel 应用， 并打开文件

        :param filename:  文件名
        :param visible:  是否显示应用界面
        :return:  返回 (app, document)
        """
        return Win32._open('Excel.Application|KET.Application', 'Workbooks', filename, visible)

    @staticmethod
    def open_powerpoint(filename, visible=True):
        """
        启动 PowerPoint 应用， 并打开文件

        :param filename:  文件名
        :param visible:  是否显示应用界面
        :return:  返回 (app, document)
        """
        return Win32._open('PowerPoint1111.Application|KWPP.Application', 'Presentations', filename, visible)

    @staticmethod
    def open_pdf(filename, visible=True):
        """
        启动 Word应用， 并打开文件

        :param filename:  文件名
        :param visible:  是否显示应用界面
        :return:  返回 (app, document)
        """
        # https://cloud.tencent.com/developer/article/1907286
        return Win32._open('Kwps.Application', 'Documents', filename, visible)

    @staticmethod
    def close(app, file_obj=None):
        """ 关闭文件， 关闭应用 """
        # noinspection PyBroadException
        try:
            if file_obj:
                if hasattr(file_obj, 'Saved'):
                    file_obj.Saved = True
                file_obj.Close()  # 关闭文档
        except Exception:
            pass

        # noinspection PyBroadException
        try:
            if app:
                app.Quit()
        except Exception:
            pass

    @staticmethod
    def find_window(title):
        """
        查找标题包含title字符串的窗口

        :param title: 窗口标题
        :return:  返回 HWND. 失败返回 0
        """
        # https://www.programcreek.com/python/example/10639/win32gui.EnumWindows
        def window_enumeration_handler(hwnd, windows):
            """Add window title and ID to array."""
            windows.append((hwnd, win32gui.GetWindowText(hwnd)))

        top_windows = []
        win32gui.EnumWindows(window_enumeration_handler, top_windows)

        for window in top_windows:
            if window[1].lower().find(title) >= 0:
                return window[0]

        return 0

    @staticmethod
    def bring_to_front(win_app):
        """ 将 app 窗口置于最前 """
        # noinspection PyBroadException
        try:
            if isinstance(win_app, CDispatch):
                title = win_app.Caption
            elif isinstance(win_app, str):
                title = win_app
            else:
                raise ValueError("Expect COM object or str, but %s is found" % repr(type(win_app)))

            hwnd = Win32.find_window(title)
            if hwnd != 0:
                win32gui.SetForegroundWindow(hwnd)
                return True
        except Exception:
            pass

        return False



from distutils.core import setup
from setuptools import find_packages

# tutor: https://developer.aliyun.com/article/778803
# license : https://choosealicense.com/community/
# download statistics https://pepy.tech/


def read_file(filename, encoding='utf-8'):
    with open(filename, "rt", encoding=encoding) as f:
        return f.read()


def read_file_lines(filename, encoding='utf-8'):
    text = read_file(filename, encoding=encoding)
    result = []
    for line in text.splitlines():
        line = line.strip()
        if line and not line.startswith('#'):
            result.append(line)
    return result


setup(name='jojo-office',  # 包的分发名称
      version='0.2.5',  # PyPI上只允许一个版本存在，如果后续代码有了任何更改，再次上传需要增加版本号
      description='Process .docx, .xlsx, .pptx, .pdf files for office automation',  # 项目的简短描述
      long_description=read_file("README.rst"),  # 详细描述，会显示在PyPI的项目描述页面。必须是rst(reStructuredText) 格式的
      author='JoStudio',
      author_email='jostudio@189.cn',
      url='https://www.jostudio.com.cn/python',
      # 项目依赖哪些库，这些库会在pip install的时候自动安装
      # install_requires=[],
      # 从文件中读取依赖库
      install_requires=read_file_lines("office/requirements.txt"),
      license='Apache License 2.0',
      packages=find_packages(),
      platforms=["all"],
      classifiers=[
          'Intended Audience :: Developers',
          'Operating System :: OS Independent',
          'Natural Language :: English',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Topic :: Software Development :: Libraries'
      ],
      )



import inspect
from types import MethodType


class SuperObject:
    """
    SuperObject

    support dynamically add methods to base object.

    """

    @staticmethod
    def _is_function_of(func, clz):
        """ whether func is a function defined in cls """
        if hasattr(func, '__qualname__'):
            qualname = func.__qualname__
            if qualname.find(clz.__name__ + '.') == 0:
                if qualname.find('_subclass') < 0:
                    return True
        return False

    @staticmethod
    def merge_methods(dst_obj, src_class):
        """
        read the methods of the src class, add to the object

        :param dst_obj:   the object which need to add method
        :param src_class:  a class which provide methods
        :return: return modified object
        """
        if dst_obj is None or src_class is None:
            return dst_obj

        if inspect.isclass(src_class):
            # analysis each attribute of the src_class
            for name, attr in inspect.getmembers(src_class):
                if isinstance(attr, property) and SuperObject._is_function_of(attr.fget, src_class):
                    # when a property of extend_class is met
                    # dynamically add property getter function as a method to the object
                    if attr.fget and not hasattr(dst_obj, 'get_' + name):
                        setattr(dst_obj, 'get_' + name, MethodType(attr.fget, dst_obj))
                    # dynamically add property setter function as a method to the object
                    if attr.fset and not hasattr(dst_obj, 'set_' + name):
                        setattr(dst_obj, 'set_' + name, MethodType(attr.fset, dst_obj))
                elif inspect.isfunction(attr) and SuperObject._is_function_of(attr, src_class):
                    # when a method is met,  dynamically add method to obj dynamically
                    setattr(dst_obj, name, MethodType(attr, dst_obj))
            return dst_obj
        else:
            raise ValueError('expect class type but %s is found' % repr(type(src_class)))

    def __init__(self, base_obj=None):
        self._methods = {}
        self._properties = {}
        if not hasattr(self, 'base_obj') or base_obj:
            # print('set base_obj')
            self.base_obj = base_obj
            if base_obj is not None:
                self._build_methods()

    def __getattribute__(self, attr):
        """ if self has the attr, return it. else if base_obj has the attr, return it."""

        # get attribute from self
        # http://www.sefidian.com/2021/06/06/python-__getattr__-and-__getattribute__-magic-methods/
        __dict__ = super().__getattribute__('__dict__')
        if attr in __dict__:
            return super().__getattribute__(attr)

        # get attribute from _methods
        if '_methods' in __dict__:
            methods = super().__getattribute__('_methods')
            if methods and attr in methods:
                m = methods[attr]
                return m

        # get attribute from base_obj
        if 'base_obj' in __dict__:
            if hasattr(super().__getattribute__('base_obj'), attr):
                return getattr(super().__getattribute__('base_obj'), attr)

        return super().__getattribute__(attr)

    def __setattr__(self, attr, value):
        """ if base_obj has the attr, set base_obj value. else if self has the attr, set its value."""
        # get dict of self
        __dict__ = super().__getattribute__('__dict__')

        # if value is a function, add it as method
        if inspect.isfunction(value):
            return self.add_method(attr, value)

        # if base_obj has the attribute, set the value to attribute of the base_obj
        if 'base_obj' in __dict__:
            if hasattr(super().__getattribute__('base_obj'), attr):
                setattr(super().__getattribute__('base_obj'), attr, value)
                return

        # set the value of the attribute of self
        super().__setattr__(attr, value)

    def _build_methods(self):
        # get method of base_obj
        if self.base_obj is not None:
            for name, attr in inspect.getmembers(self.base_obj):
                if name.startswith('__'):
                    continue
                if inspect.ismethod(attr) or inspect.isbuiltin(attr):
                    self._methods[name] = attr

        # get method of self
        for name, attr in inspect.getmembers(type(self)):
            if name.startswith('__'):
                continue
            if inspect.isfunction(attr) or inspect.isbuiltin(attr):
                self._methods[name] = MethodType(attr, self)

    @property
    def methods(self):
        """ return the methods """
        if not self._methods:
            self._build_methods()
        return self._methods

    @property
    def properties(self):
        """ return the properties """
        if not self._properties:
            # get properties of self
            for name, attr in inspect.getmembers(type(self)):
                if name.startswith('__'):
                    continue
                if isinstance(attr, property):
                    if name not in ['methods', 'properties', 'attributes']:
                        self._properties[name] = attr

            # get properties of base_obj
            if self.base_obj is not None:
                for name, attr in inspect.getmembers(type(self.base_obj)):
                    if name.startswith('__'):
                        continue
                    if isinstance(attr, property):
                        self._properties[name] = attr

        return self._properties

    @property
    def attributes(self):
        """ return the attributes """
        result = {}

        # get attrs of self
        for name, attr in inspect.getmembers(self):
            if name.startswith('__'):
                continue
            if inspect.isfunction(attr) or inspect.ismethod(attr) or inspect.isbuiltin(attr):
                continue
            result[name] = ''

        # get attrs of base_obj
        if self.base_obj is not None:
            for name, attr in inspect.getmembers(self.base_obj):
                if name.startswith('__'):
                    continue
                if inspect.isfunction(attr) or inspect.ismethod(attr) or inspect.isbuiltin(attr):
                    continue
                result[name] = attr

        return result

    def add_method(self, name, func=None):
        """ dynamically add a method """
        if isinstance(name, str):
            if inspect.isfunction(func):
                m = self.methods
                m[name] = MethodType(func, self)
        elif inspect.isclass(name):
            src_class = name
            # read the methods of the class
            for name, attr in inspect.getmembers(src_class):
                if isinstance(attr, property) and SuperObject._is_function_of(attr.fget, src_class):
                    # when a property of extend_class is met
                    if attr.fget:
                        self.add_method('get_' + name, attr.fget)
                    if attr.fset:
                        self.add_method('set_' + name, attr.fset)
                elif inspect.isfunction(attr) and SuperObject._is_function_of(attr, src_class):
                    # when a method is met
                    self.add_method(name, attr)

        return self



from urllib.parse import quote, urlencode
from urllib.request import Request, urlopen
import json


__version__ = "0.5"
__author__ = "JoStudio"


# UserAgent
class UserAgent:
    chrome = 'Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/31.0.1650.16 Safari/537.36'
    default = 'Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/31.0.1650.16 Safari/537.36'


# Http V0.5
class Http:
    """
    Http Class

    """
    @staticmethod
    def get(url, params=None, headers=None):
        """ send a GET request"""
        return Http().request(url, params, headers=headers)

    @staticmethod
    def post(url, data=None, headers=None):
        """ send a POST request"""
        return Http().request(url, data, headers=headers, method='POST')

    @staticmethod
    def delete(url, params=None, headers=None):
        """ send a DELETE request"""
        return Http().request(url, params, headers=headers, method='DELETE')

    @staticmethod
    def head(url, params=None, headers=None):
        """ send a HEAD request"""
        return Http().request(url, params, headers=headers, method='HEAD')

    @staticmethod
    def put(url, data=None, headers=None):
        """ send a get request"""
        return Http().request(url, data, headers=headers, method='PUT')

    def __init__(self):
        self.req = None
        self.response = None
        self._content = None

    @staticmethod
    def _compose_get_url(url, params):
        """ compose HTTP GET url"""
        if isinstance(params, dict):
            if url.find('?') < 0:
                url += '?'
            else:
                if not url.endswith('&'):
                    url += '&'
            for name in params:
                url += quote(name) + '=' + quote(params[name])
        elif isinstance(params, list) or isinstance(params, tuple):
            for index, val in enumerate(params):
                url = url.replace('{%s}' % index, quote(str(val)))
        elif isinstance(params, str):
            url = url.replace('{0}', quote(params))
        elif params is None:
            pass
        else:
            raise ValueError('invalid params type %s' % params)

        return url

    def request(self, url, data=None, headers=None, method='GET'):
        r"""Sends a request.

            :param url: URL for the new :class:`Request` object.
            :param data: (optional) Dictionary, list of tuples or bytes to send
                in the query string for the :class:`Request`.
            :param headers: Optional headers of the request.
            :param method: Optional method of the request, 'GET', 'POST', 'PUT', 'DELETE', 'HEAD'.
            :return: return self
            """
        self.req = None
        self.response = None
        self._content = None
        # noinspection PyBroadException
        if method in ['GET', 'HEAD', 'DELETE']:
            url = Http._compose_get_url(url, data)
            data = None
        headers = {} if headers is None else headers
        req_headers = {"User-Agent": UserAgent.default}
        if isinstance(headers, dict):
            for key in headers:
                req_headers[key] = headers[key]
        self.req = Request(url, data=data, headers=headers, method="GET")
        self.response = urlopen(self.req)
        return self

    def save_to_file(self, filename):
        """ save response to file """
        filename = str(filename)
        if filename.rfind('.') < 0:
            filename += self._get_file_ext()
        f = open(filename, 'wb')
        f.write(self.response.read())
        f.close()
        return filename

    def _get_file_ext(self):
        fmt = self.response.headers.get_content_type()
        if isinstance(fmt, str):
            if fmt.find('/') >= 0:
                fmt = fmt[fmt.find('/') + 1:]
                return '.' + fmt
        return ''

    def header(self, name, sub_name=None):
        if self.response:
            name = name.lower()
            for key in self.response.headers:
                if key.lower() == name:
                    val = self.response.headers[key]
                    vals = val.split(';')
                    if sub_name is None:
                        if len(vals) > 0:
                            return vals[0].strip()
                    else:
                        for item in vals:
                            pos = item.find('=')
                            if pos > 0:
                                name = item[:pos]
                                value = item[pos+1:]
                                if name.lower() == sub_name.lower():
                                    return value

    @property
    def content_type(self):
        if self.response:
            return self.response.headers.get_content_type()

    @property
    def status_code(self):
        if self.response:
            return self.response.status
        return 0

    @property
    def text(self):
        """ get the text of response """
        if self.response:
            charset = self.header('Content-Type', 'charset')
            charset = 'utf-8' if not charset else charset
            return str(self.response.read(), encoding=charset)

    @property
    def content(self):
        """ get the bytes of response"""
        if not self._content:
            if self.response:
                self._content = self.response.read()
        return self._content

    def json(self, **kwargs):
        """ get the json data of response """
        return json.loads(self.text, **kwargs)


